/*  $Id: ontodella_server.pl,v 1.11 2004/01/15 16:55:44 kimvilja Exp $

    OnToDella prolog httpd server

    Author:  Arttu Valo
    E-mail:  arttu.valo@helsinki.fi
    WWW:     http://www.cs.helsinki.fi/group/seco/fmsweb

Copyright (c) 2003-2004 Helsinki Institute for Information Technology

Permission is hereby granted, free of charge, to any person obtaining 
a copy of this software and associated documentation files (the "Software"), 
to deal in the Software without restriction, including without limitation 
the rights to use, copy, modify, merge, publish, distribute, sublicense, 
and/or sell copies of the Software, and to permit persons to whom the 
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included 
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
THE SOFTWARE.
*/

:- initialization
    mutex_create( logfilemutex ).  % mutex for log file access (writing)


% Provide logging predicate for all threads (prolog should internally
% handle thread-synchronization!? at least so the docs claimed.)
%:- dynamic
%    write_log_message/1.

:- load_files([library('http/thread_httpd')], [ silent(true) ]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% These predicates must be located here, before calling them... :-o

find_ontodella_home('.'). % use the current directory

find_ontodella_home(Home):-
    ontodella_home(Home).

ontodella_home(Home) :- % the given directory
    current_prolog_flag(argv, Args), 
    member(Arg, Args),
    string_concat('home=', Home, Arg).

ontodella_file( Path, Path ) :-
    exists_file( Path ).

ontodella_file( Filename, PathA ) :-
    find_ontodella_home( Home ),
    sformat( Path, "~w/~w", [Home, Filename] ),
    string_to_atom( Path, PathA ),
    exists_file( PathA ) ->
%	format( 'FILE: ~w~n', [PathA] )
       true;
    format( 'ERROR: ontodella_file/2: NO SUCH FILE: ~w => ~w~n', [Filename, PathA] ),
    fail.

ontodella_config_file( File ) :-
    current_prolog_flag(argv, Args), 
    member(Arg, Args),
    string_concat('config=', File, Arg).

ontodella_config_file( File ) :-
    ontodella_file( 'conf/ontodella_config.pl', File ).

ontodella_httpd_port(Port) :- % returns the port
    current_prolog_flag(argv, Args), 
    member(Arg, Args),
    string_concat('port=', PortS, Arg),
    string_to_atom(PortS, PortA),
    (
        PortA='none',
        Port='none';
        ontodella_request_handler:atom_to_integer(PortA, Port)
    ),
    !.

ontodella_httpd_port(Port) :-
    get_config_http_port(Port).

ontodella_log_file(File) :-
    current_prolog_flag(argv, Args), 
    member(Arg, Args),
    string_concat('log=', FileS, Arg),
    string_to_atom(FileS, File),
    !.

ontodella_log_file(File) :-
    get_config_logger_file(File).

ontodella_testfile(File) :-
    current_prolog_flag(argv, Args), 
    member(Arg, Args),
    string_concat('test_file=', FileS, Arg),
    string_to_atom(FileS, File),
    !.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


:- 
        ontodella_config_file( OntConf ),
	[OntConf],

	ontodella_file( 'src/prolog/ontodella_request_handler.pl', ReqHandler ),
	[ReqHandler].


%server
server :-
    write('Ontodella server starting procedure begins.\n'),
    write('0. Checking environment etc.\n'),
    !,
    ontodella_httpd_port( Port ),
format('httpd port ~w ~n', [Port]),
    ( 
        Port \= 'none' ->
           tcp_port_used( Port ),
           format('httpd port ~w ok.~n', [Port]);
        format('Just load everything, but do not start the server.~n')
    ),
    (   % if logging is requested, enable it
        get_config_use_logging,
        start_httpd_logger,
        % give the other thread some time to get started
        sleep( 5 );

        true
    ),

    write('1. Loading data.\n'),
    load_rdfs_files,

    write('2. Running tests.\n'),
    (
        ontodella_testfile(TestFile),
        [TestFile],
        ontodella_test(ErrorCounter),
	(
	    ErrorCounter > 0,
	    format('Got ~w warning(s)!~n', [ErrorCounter]);
	    true
	);
	true
    ),

    write('3. Preprocessing data.\n'),
    catch(ontodella_preprocess,PreExcp, format('When doing preprocessing, following exception occured: ~w~n', [PreExcp])), % if possible, do preprocessing

    write('4. Starting the server.\n'),

    (
        Port \= 'none',
        (
             server( Port, [workers(10)] ),
             writef( 'OntoDella Prolog Server running on port %d.\n', [Port] );

             writef( 'There was a problem starting the server!' ),
             stop_server
        );
        format('port=none = do not start the server.')
    ).


%server( +Port, +Options)
server( Port, Options ) :-
    writef( 'Starting OntoDella Prolog HTTPD on port %d with options %w ...\n', [Port, Options] ),
    http_server(reply,
		[ port(Port),
		  timeout(20)
		  | Options
		]).


%load_rdfs_files
load_rdfs_files :-
    get_config_rdfs_files( RdfsFiles ),
    load_rdfs_files( RdfsFiles ).

%load_rdfs_files( [+RdfsFiles] )
load_rdfs_files( RdfsFiles ) :-
    writeln( 'Loading RDF(S) files ...' ),
    % if we cannot load the RDF(S) files, refuse to go on
    checklist( server_load_rdf, RdfsFiles ),
    writeln( 'Loaded RDF(S) files.' ).


% Loads an RDF(S) file into database
%load_rdf( +Filename )

% Try to guess the file path.

server_load_rdf( Filename2 ) :-
    ontodella_file( Filename2, Filename ),
    writef( 'Loading %w\n', [Filename] ),
    (
        load_rdf( Filename, H ),
        string_concat( 'Loaded ', Filename, FilenameLoaded ),
        writef( '%w\n', [FilenameLoaded] ),
        (
            get_config_use_logging,
            thread_send_message( httpdlogger, FilenameLoaded );

            true
        ),
        checklist( assert, H ),
%        writeln( 'checklist-asserted all facts'),
        (
            get_config_use_logging,
            thread_send_message( httpdlogger, 'Loading completed' );

            true
        );

        (
            get_config_use_logging,
            thread_send_message( httpdlogger, 'Loading FAILED!' );

            true
        ),
        fail
    ).


%unload_rdfs_files
unload_rdfs_files :-
    retractall( rdf( _, _, _ ) ),
    (
        get_config_use_logging,
        thread_send_message( httpdlogger, 'All RDF(S) facts retracted. Database empty.' );

        true
    ).


%start_httpd_logger
start_httpd_logger :-
%    catch( thread_create( log_listener_loop, Id, [alias( httpdlogger )] ),
%        stop_server, _ ),
    (
      ( ontodella_log_file( LogFile );
        LogFile='prolog-httpd-server.log'
      ),
      ( ontodella_home( Home ),
        sformat( LogFilePath, '~w/logs/~w', [Home, LogFile]);
        LogFilePath=LogFile
      )
    ),
    thread_create( log_listener_loop( LogFilePath ), _,
        [alias(httpdlogger)] ),
    thread_send_message( httpdlogger, 'Server started.' ).


% main logger loop
%log_listener_loop( +Filename )
log_listener_loop( Filename ) :-
    mutex_lock( logfilemutex ),
    open( Filename, append, _, [alias(httpdlogoutput), buffer(line)] ),
    current_output( OldStream ),
    set_output( httpdlogoutput ),
    writeln( '===========' ),
    get_time( TimeFloat ),
    convert_time( TimeFloat, Time ),
    writef( 'Prolog HTTP Server started at %w\n', [Time] ),
    set_output( OldStream ),
    mutex_unlock( logfilemutex ),
    % Below: the loop
    repeat,  % eternal loop!? (at least it should be until stop_httpd_logger is called)
    thread_get_message( Message ),
    write_log_message( Message ),
    Message = 'stoplogging' ->(
        write( user_output, 'Stopping httpd logger.\n' ),
        stop_httpd_logger
    ),
    fail,  % fail -> backtrack and repeat
    !.

% stops the httpd logger. DO NOT CALL DIRECTLY! Instead, send a message to this
% thread like this:
% thread_send_message( httpdlogger, 'stoplogging' ).
%stop_http_logger
stop_httpd_logger :-
    get_time( TimeFloat ),
    convert_time( TimeFloat, Time ),
    mutex_lock( logfilemutex ),
    current_output( OldStream ),
    set_output( httpdlogoutput ),
    writef( 'Prolog HTTP Server stopped at %w\n', [Time] ),
    close( httpdlogoutput, [force(true)] ),
    set_output( OldStream ),
    mutex_unlock( logfilemutex ),
    thread_exit( true ).

%write_log_message( +Message )
write_log_message( Message ) :-
    get_time( TimeFloat ),
    convert_time( TimeFloat, Time ),
    mutex_lock( logfilemutex ),
    current_output( OldStream ),
    set_output( httpdlogoutput ),
    writef( '[%w] %w\n', [Time, Message] ),
    flush_output( httpdlogoutput ),
    set_output( OldStream ),
    mutex_unlock( logfilemutex ).

%stop_server
stop_server :-
    (
        get_config_use_logging,
        thread_send_message( httpdlogger, 'stoplogging' ),
        sleep( 5 ),
        thread_signal( httpdlogger, halt );  % "insurance"

        true
    ).
%    stop_httpd_logger,
%    threads,
%    halt. % we do not want to halt the prolog process.


%is the port already in use?
%tcp_port_used( +Port )
tcp_port_used( Port ) :-
   catch((
 	tcp_socket(S), 
	tcp_bind(S, Port), 
	tcp_close_socket(S)
	),
	Exception,
	(
	  format('Port ~w is already in use! Exception: ~w~n', [Port, Exception]),
 	  fail
	)),
	true.
